use crate::api;
use crate::core;
use crate::repositories;
use diesel::{
  r2d2::{ConnectionManager, Pool},
  PgConnection,
};

///
pub struct Synchronizer {
  pool: Pool<ConnectionManager<PgConnection>>,
  api_config: core::config::ApiConfig,
  synchronization_history_repository: repositories::synchronization_history::repository::Repository,
  betting_history_repository: repositories::betting_history::repository::Repository,
  betting_api: api::betting::api::Api,
}

impl std::fmt::Debug for Synchronizer {
  fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
    f.debug_struct("Synchronizer of api.kgon.identity").finish()
  }
}

impl std::clone::Clone for Synchronizer {
  fn clone(&self) -> Self {
    Self {
      pool: self.pool.clone(),
      api_config: self.api_config.clone(),
      synchronization_history_repository:
        repositories::synchronization_history::repository::Repository::new(),
      betting_history_repository: repositories::betting_history::repository::Repository::new(),
      betting_api: api::betting::api::Api::new(self.api_config.clone()),
    }
  }
}

impl Synchronizer {
  ///
  pub fn new(
    pool: Pool<ConnectionManager<PgConnection>>,
    api_config: core::config::ApiConfig,
  ) -> Synchronizer {
    Synchronizer {
      pool,
      api_config: api_config.clone(),
      synchronization_history_repository:
        repositories::synchronization_history::repository::Repository::new(),
      betting_history_repository: repositories::betting_history::repository::Repository::new(),
      betting_api: api::betting::api::Api::new(api_config),
    }
  }

  pub async fn betting_history(&self) -> Result<(), Box<dyn std::error::Error>> {
    let start_at = (chrono::Utc::now()).timestamp();

    if let Err(e) = async {
      let conn = match self.pool.get() {
        Ok(c) => c,
        Err(e) => {
          return Err(api::core::models::Error {
            code: -1,
            msg: Some(format!("pool.get error: {}", e)),
          });
        }
      };

      let sdate = match self
        .synchronization_history_repository
        .select_latest_by_item(
          &conn,
          repositories::synchronization::models::ITEM_BETTING_HISTORY.to_string(),
        ) {
        Ok(v) => match v {
          Some(s) => s.data,
          None => None,
        },
        Err(e) => {
          return Err(api::core::models::Error {
            code: -1,
            msg: Some(format!(
              "synchronization_history_repository.select_latest_by_item error: {}",
              e
            )),
          });
        }
      };

      let req = api::betting::models::ListBettingsRequest {
        vendor_key: None,
        sdate,
        edate: None,
        group_key: None,
        username: None,
        limit: 2000,
      };
      let res = match self.betting_api.list_bettings(req).await {
        Ok(v) => v,
        Err(e) => {
          return Err(api::core::models::Error {
            code: -1,
            msg: Some(format!("betting_api.list_bettings error: {:?}", e)),
          });
        }
      };

      let last_object_id = res.last_object_id;
      let mut last_utc_created_at: Option<String> = None;
      let mut new_betting_history: Vec<repositories::betting_history::models::NewBettingHistory> =
        vec![];

      for b in res.transactions {
        if let Some(loi) = &last_object_id {
          if b._id.eq(loi) {
            last_utc_created_at = Some(b.utc_created_at.clone());
          }
        }

        let created_at =
          match chrono::NaiveDateTime::parse_from_str(&b.created_at, "%Y-%m-%d %H:%M:%S") {
            Ok(t) => t.timestamp(),
            Err(e) => {
              return Err(api::core::models::Error {
                code: -1,
                msg: Some(format!(
                  "chrono::NaiveDateTime::parse_from_str error: {}",
                  e
                )),
              });
            }
          };
        let utc_created_at = match chrono::DateTime::parse_from_rfc3339(b.utc_created_at.as_str()) {
          Ok(t) => t.timestamp(),
          Err(e) => {
            return Err(api::core::models::Error {
              code: -1,
              msg: Some(format!("chrono::DateTime::parse_from_rfc3339 error: {}", e)),
            });
          }
        };
        let detail = b.detail.map(|v| v.get().to_string());

        new_betting_history.push(repositories::betting_history::models::NewBettingHistory {
          id: b._id,
          vendor_id: b.vendor_id,
          vendor_name: b.vendor_name,
          game_id: b.game_id,
          game_name: b.game_name,
          game_category: b.game_category,
          game_type: b.game_type,
          currency: b.currency,
          cash: b.cash,
          before_cash: b.before_cash,
          after_cash: b.after_cash,
          group_key: b.group_key,
          detail,
          is_bonus: b.is_bonus,
          is_promo: b.is_promo,
          is_jackpot: b.is_jackpot,
          site_username: b.site_username,
          key: b.key,
          ref_id: b.ref_id,
          o_ref_id: b.o_ref_id,
          betting_type: b.betting_type,
          category: b.category,
          created_at,
          utc_created_at,
        });
      }

      if !new_betting_history.is_empty() {
        match self
          .betting_history_repository
          .inserts(&conn, &new_betting_history)
        {
          Ok(v) => v,
          Err(e) => {
            return Err(api::core::models::Error {
              code: -1,
              msg: Some(format!("betting_history_repository.inserts error: {:?}", e)),
            });
          }
        };
      }

      if let Err(e) = self.synchronization_history_repository.insert(
        &conn,
        &repositories::synchronization_history::models::NewSynchronizationHistory {
          item: repositories::synchronization::models::ITEM_BETTING_HISTORY.to_string(),
          start_at,
          complete_at: (chrono::Utc::now()).timestamp(),
          code: 0,
          message: None,
          data: last_utc_created_at,
        },
      ) {
        println!("synchronization_history.insert error: {}", e);
      };

      Ok::<(), api::core::models::Error>(())
    }
    .await
    {
      let conn = match self.pool.get() {
        Ok(c) => c,
        Err(e) => {
          println!("pool.get error: {}", e);
          return Ok(());
        }
      };

      if let Err(e) = self.synchronization_history_repository.insert(
        &conn,
        &repositories::synchronization_history::models::NewSynchronizationHistory {
          item: repositories::synchronization::models::ITEM_BETTING_HISTORY.to_string(),
          start_at,
          complete_at: (chrono::Utc::now()).timestamp(),
          code: e.code,
          message: e.msg,
          data: None,
        },
      ) {
        println!("synchronization_history.insert error: {}", e);
      };
    }

    Ok(())
  }
}
